
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

df_filtered = pd.read_csv('A_greater_B.csv')

import numpy as np
import pandas as pd
import glob
from model.ctabgan import CTABGAN

synthesizer =  CTABGAN(df_filtered,
                 test_ratio = 0.20,
                 categorical_columns = [], 
                 log_columns = [],
                 mixed_columns= {},
                 general_columns = [],
                 non_categorical_columns = [],
                 integer_columns = [],
                 problem_type= None) 

synthesizer.fit(30)

synthesizer.generate_samples(30)


"""
Generative model training algorithm based on the CTABGANSynthesiser

"""
import pandas as pd
import time
from .pipeline.data_type_assigner import Data_type_assigner
from .pipeline.data_preparation import DataPrep
from .pipeline.Column_assigner import Column_assigner, Transform_type
from .synthesizer.ctabgan_synthesizer import CTABGANSynthesizer


import warnings
import numpy as np

warnings.filterwarnings("ignore")

class CTAB_XTRA_DP():

    def __init__(self,
                 df,
                 test_ratio = 0.20,
                 categorical_columns = [], 
                 log_columns = [],
                 mixed_columns= {},
                 gaussian_columns = [],
                 non_categorical_columns = [],
                 integer_columns = [],
                 problem_type = ("Classification",'workclass'),
                 dp_constraints = {
                    "epsilon_budget": 10,
                    "delta": None,
                    "sigma": None,
                    "clip_coeff": 1
                 }
                 ):

        self.__name__ = 'CTAB_XTRA_DP'
              
        self.synthesizer = CTABGANSynthesizer()
        self.raw_df = df
        self.test_ratio = test_ratio
        self.categorical_columns = categorical_columns
        self.log_columns = log_columns
        self.mixed_columns = mixed_columns
        self.gaussian_columns = gaussian_columns
        self.integer_columns = integer_columns

        self.problem_type = problem_type
        self.dp_constraints = dp_constraints

                
    def fit(self,epochs = 100):
        
        start_time = time.time()
        
     
        self.data_type_assigner = Data_type_assigner(self.raw_df, self.integer_columns)


       

        self.raw_df = self.data_type_assigner.assign(self.raw_df)

        self.data_prep = DataPrep(self.raw_df, self.categorical_columns, self.log_columns)

        self.prepared_data = self.data_prep.preprocesses_transform(self.raw_df)
        


        self.synthesizer.fit(self.prepared_data , self.data_prep, self.dp_constraints, self.categorical_columns, self.mixed_columns, self.gaussian_columns, self.problem_type,epochs)
        return
        


    def generate_samples(self,n=100,conditioning_column = None,conditioning_value = None):
        column_index = None
        column_value_index = None
        if conditioning_column and conditioning_value:
            column_index = self.prepared_data.columns.get_loc(conditioning_column) if conditioning_column in self.prepared_data.columns else ValueError("Conditioning column", conditioning_column, "not found in the data columns")
            column_value_index = self.data_prep.get_label_encoded(column_index, conditioning_value)

        sample_transformed = self.synthesizer.sample(n, column_index, column_value_index)
        sample_transformed = pd.DataFrame(sample_transformed, columns=self.prepared_data.columns)
        
        sample = self.data_prep.preprocesses_inverse_transform(sample_transformed)
        sample_with_data_types = self.data_type_assigner.assign(sample)
        return sample_with_data_types
        
        
  

    def generate_samples_index(self,n=100,index=None):

        sample = self.synthesizer.sample(n,0,index)
        sample_df = self.data_prep.inverse_prep(sample)

        return sample_df



import numpy as np
import pandas as pd
import glob
from model.ctabgan import CTABGAN
df = pd.read_csv("Real_Datasets/car.csv")
df = df.drop(columns=['Year','Model'])

synthesizer =  CTABGAN(df,
                 test_ratio = 0.20,
                 categorical_columns = ["Brand","Model","Fuel_Type","Transmission"], 
                 log_columns = ["Mileage"],
                 mixed_columns= {},
                 gaussian_columns = [],
                 non_categorical_columns = [],
                 integer_columns = ["Price"],
                 problem_type= {"Classification": 'Price'}) 

synthesizer.fit(1)



import warnings

class Data_type_assigner:
    def __init__(self, data,integer_columns = []):
    
        self.data_type = data.dtypes
        self.integer_columns = integer_columns + data.select_dtypes(include=['int64']).columns.tolist()
        for column in integer_columns:
            if column not in self.data_type: 
                warnings.warn(f"Column {column} not found in data, ignoring integer assignment")
                continue
            if data[column].isna().any(): continue # If we have nan values, we cant assign the column to interger, however we still treat it as integer
            data[column] = data[column].astype(int)
        
        self.number_of_decimals = self.get_column_desimal(data)
        

        

    def assign(self, data):
        for i in range(len(data.columns)):
            if self.number_of_decimals[i] is not None:
                data.iloc[:, i] = data.iloc[:, i].round(self.number_of_decimals[i])
      
        data = data.astype(self.data_type)
        return data

    
    def get_column_desimal(self, data):
        decimals = [None] * len(data.columns)
        for i in range(len(data.columns)):
            if data.iloc[:, i].dtype == 'float64':
                decimals[i] = data.iloc[:, i].apply(lambda x: len(str(x).split('.')[1]) if '.' in str(x) else 0).max()

        for column in self.integer_columns: # The integers columns have 0 decimals by default
            idx = data.columns.get_loc(column)
            decimals[idx] = 0
        return decimals



import numpy as np
import pandas as pd
import glob
from model.ctabgan import CTABGAN
df = pd.read_csv("Real_Datasets/adult.csv")
#df = df.drop(columns=['Year','Model'])
df = df[['age', 'workclass', 'capital-gain']]
df = df.head(1000)
df.loc[:9, 'capital-gain'] = np.nan

synthesizer =  CTABGAN(df,
                 test_ratio = 0.20,
                 categorical_columns = ["workclass"], 
                 log_columns = {}, # Fuck this log, task for tommorrow
                 mixed_columns= {"capital-gain": [0,np.nan]},
                 gaussian_columns = ["age"],
                 non_categorical_columns = [],
                 integer_columns = ["age","capital-gain"]) 


synthesizer.fit(1)


from setuptools import setup, find_packages

setup(
    name="ctab_xtra_dp",
    version="0.2.0",
    packages=find_packages(),
    install_requires=[],  # List dependencies here, e.g., ["numpy", "requests"]
    description="A sample Python package",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="kem0sabe",
    author_email="martivl@stud.ntnu.no",
    url="https://github.com/Kem0sabe/Package_example",  # Update with your GitHub or website
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy as np
import pandas as pd 
from sklearn import metrics
from sklearn import model_selection
from sklearn.preprocessing import MinMaxScaler,StandardScaler
from sklearn.neural_network import MLPClassifier
from sklearn.linear_model import LogisticRegression, LinearRegression, Ridge, Lasso, BayesianRidge
from sklearn import svm,tree
from sklearn.ensemble import RandomForestClassifier
from dython.nominal import associations
from scipy.stats import wasserstein_distance
from scipy.spatial import distance
import warnings

warnings.filterwarnings("ignore")

def supervised_model_training(x_train, y_train, x_test, 
                              y_test, model_name,problem_type):
  
  if model_name == 'lr':
    model  = LogisticRegression(random_state=42,max_iter=500) 
  elif model_name == 'svm':
    model  = svm.SVC(random_state=42,probability=True)
  elif model_name == 'dt':
    model  = tree.DecisionTreeClassifier(random_state=42)
  elif model_name == 'rf':      
    model = RandomForestClassifier(random_state=42)
  elif model_name == "mlp":
    model = MLPClassifier(random_state=42,max_iter=100)
  elif model_name == "l_reg":
    model = LinearRegression()
  elif model_name == "ridge":
    model = Ridge(random_state=42)
  elif model_name == "lasso":
    model = Lasso(random_state=42)
  elif model_name == "B_ridge":
    model = BayesianRidge()
  
  model.fit(x_train, y_train)
  pred = model.predict(x_test)

  if problem_type == "Classification":
    if len(np.unique(y_train))>2:
      predict = model.predict_proba(x_test)        
      acc = metrics.accuracy_score(y_test,pred)*100
      auc = metrics.roc_auc_score(y_test, predict,average="weighted",multi_class="ovr")
      f1_score = metrics.precision_recall_fscore_support(y_test, pred,average="weighted")[2]
      return [acc, auc, f1_score] 

    else:
      predict = model.predict_proba(x_test)[:,1]    
      acc = metrics.accuracy_score(y_test,pred)*100
      auc = metrics.roc_auc_score(y_test, predict)
      f1_score = metrics.precision_recall_fscore_support(y_test,pred)[2].mean()
      return [acc, auc, f1_score] 
  
  else:
    mse = metrics.mean_absolute_percentage_error(y_test,pred)
    evs = metrics.explained_variance_score(y_test, pred)
    r2_score = metrics.r2_score(y_test,pred)
    return [mse, evs, r2_score]


#TODO: Fix to work with pandas df instead of path
def get_utility_metrics(data_real,fake_paths,scaler="MinMax",type={"Classification":["lr","dt","rf","mlp"]},test_ratio=.20):

    data_real = pd.read_csv(real_path).to_numpy()
    data_dim = data_real.shape[1]

    data_real_y = data_real[:,-1]
    data_real_X = data_real[:,:data_dim-1]

    problem = list(type.keys())[0]
    
    models = list(type.values())[0]
    
    if problem == "Classification":
      X_train_real, X_test_real, y_train_real, y_test_real = model_selection.train_test_split(data_real_X ,data_real_y, test_size=test_ratio, stratify=data_real_y,random_state=42) 
    else:
      X_train_real, X_test_real, y_train_real, y_test_real = model_selection.train_test_split(data_real_X ,data_real_y, test_size=test_ratio,random_state=42) 
    

    if scaler=="MinMax":
        scaler_real = MinMaxScaler()
    else:
        scaler_real = StandardScaler()
        
    scaler_real.fit(X_train_real)
    X_train_real_scaled = scaler_real.transform(X_train_real)
    X_test_real_scaled = scaler_real.transform(X_test_real)

    all_real_results = []
    for model in models:
      real_results = supervised_model_training(X_train_real_scaled,y_train_real,X_test_real_scaled,y_test_real,model,problem)
      all_real_results.append(real_results)
      
    all_fake_results_avg = []
    
    for fake_path in fake_paths:
      data_fake  = pd.read_csv(fake_path).to_numpy()
      data_fake_y = data_fake[:,-1]
      data_fake_X = data_fake[:,:data_dim-1]

      if problem=="Classification":
        X_train_fake, _ , y_train_fake, _ = model_selection.train_test_split(data_fake_X ,data_fake_y, test_size=test_ratio, stratify=data_fake_y,random_state=42) 
      else:
        X_train_fake, _ , y_train_fake, _ = model_selection.train_test_split(data_fake_X ,data_fake_y, test_size=test_ratio,random_state=42)  

      if scaler=="MinMax":
        scaler_fake = MinMaxScaler()
      else:
        scaler_fake = StandardScaler()
      
      scaler_fake.fit(data_fake_X)
      
      X_train_fake_scaled = scaler_fake.transform(X_train_fake)
      
      all_fake_results = []
      for model in models:
        fake_results = supervised_model_training(X_train_fake_scaled,y_train_fake,X_test_real_scaled,y_test_real,model,problem)
        all_fake_results.append(fake_results)

      all_fake_results_avg.append(all_fake_results)
    
    diff_results = np.array(all_real_results)- np.array(all_fake_results_avg).mean(axis=0)

    return diff_results

def stat_sim(real,fake,cat_cols=None):
    
    Stat_dict={}
    
    #real = pd.read_csv(real_path)
    #fake = pd.read_csv(fake_path)

    really = real.copy()
    fakey = fake.copy()

    real_corr = associations(real, compute_only=True)["corr"]

    fake_corr = associations(fake, compute_only=True)["corr"]

    corr_dist = np.linalg.norm(real_corr - fake_corr)
    
    cat_stat = []
    num_stat = []
    
    for column in real.columns:
        
        if column in cat_cols:

            real_pdf=(really[column].value_counts()/really[column].value_counts().sum())
            fake_pdf=(fakey[column].value_counts()/fakey[column].value_counts().sum())
            categories = (fakey[column].value_counts()/fakey[column].value_counts().sum()).keys().tolist()
            sorted_categories = sorted(categories)
            
            real_pdf_values = [] 
            fake_pdf_values = []

            for i in sorted_categories:
                real_pdf_values.append(real_pdf[i])
                fake_pdf_values.append(fake_pdf[i])
            
            if len(real_pdf)!=len(fake_pdf):
                zero_cats = set(really[column].value_counts().keys())-set(fakey[column].value_counts().keys())
                for z in zero_cats:
                    real_pdf_values.append(real_pdf[z])
                    fake_pdf_values.append(0)
            Stat_dict[column]=(distance.jensenshannon(real_pdf_values,fake_pdf_values, 2.0))
            cat_stat.append(Stat_dict[column])    
            print("column: ", column, "JSD: ", Stat_dict[column])  
        else:
            scaler = MinMaxScaler()
            scaler.fit(real[column].values.reshape(-1,1))
            l1 = scaler.transform(real[column].values.reshape(-1,1)).flatten()
            l2 = scaler.transform(fake[column].values.reshape(-1,1)).flatten()
            Stat_dict[column]= (wasserstein_distance(l1,l2))
            print("column: ", column, "WD: ", Stat_dict[column])
            num_stat.append(Stat_dict[column])

    return [np.mean(num_stat),np.mean(cat_stat),corr_dist]

def privacy_metrics(real,fake,data_percent=15):
    
    #real = pd.read_csv(real_path).drop_duplicates(keep=False)
    #fake = pd.read_csv(fake_path).drop_duplicates(keep=False)

    real_refined = real.sample(n=int(len(real)*(.01*data_percent)), random_state=42).to_numpy()
    fake_refined = fake.sample(n=int(len(fake)*(.01*data_percent)), random_state=42).to_numpy()

    scalerR = StandardScaler()
    scalerR.fit(real_refined)
    scalerF = StandardScaler()
    scalerF.fit(fake_refined)
    df_real_scaled = scalerR.transform(real_refined)
    df_fake_scaled = scalerF.transform(fake_refined)
    
    dist_rf = metrics.pairwise_distances(df_real_scaled, Y=df_fake_scaled, metric='minkowski', n_jobs=-1)
    dist_rr = metrics.pairwise_distances(df_real_scaled, Y=None, metric='minkowski', n_jobs=-1)
    rd_dist_rr = dist_rr[~np.eye(dist_rr.shape[0],dtype=bool)].reshape(dist_rr.shape[0],-1)
    dist_ff = metrics.pairwise_distances(df_fake_scaled, Y=None, metric='minkowski', n_jobs=-1)
    rd_dist_ff = dist_ff[~np.eye(dist_ff.shape[0],dtype=bool)].reshape(dist_ff.shape[0],-1) 
    smallest_two_indexes_rf = [dist_rf[i].argsort()[:2] for i in range(len(dist_rf))]
    smallest_two_rf = [dist_rf[i][smallest_two_indexes_rf[i]] for i in range(len(dist_rf))]       
    smallest_two_indexes_rr = [rd_dist_rr[i].argsort()[:2] for i in range(len(rd_dist_rr))]
    smallest_two_rr = [rd_dist_rr[i][smallest_two_indexes_rr[i]] for i in range(len(rd_dist_rr))]
    smallest_two_indexes_ff = [rd_dist_ff[i].argsort()[:2] for i in range(len(rd_dist_ff))]
    smallest_two_ff = [rd_dist_ff[i][smallest_two_indexes_ff[i]] for i in range(len(rd_dist_ff))]
    nn_ratio_rr = np.array([i[0]/i[1] for i in smallest_two_rr])
    nn_ratio_ff = np.array([i[0]/i[1] for i in smallest_two_ff])
    nn_ratio_rf = np.array([i[0]/i[1] for i in smallest_two_rf])
    nn_fifth_perc_rr = np.percentile(nn_ratio_rr,5)
    nn_fifth_perc_ff = np.percentile(nn_ratio_ff,5)
    nn_fifth_perc_rf = np.percentile(nn_ratio_rf,5)

    min_dist_rf = np.array([i[0] for i in smallest_two_rf])
    fifth_perc_rf = np.percentile(min_dist_rf,5)
    min_dist_rr = np.array([i[0] for i in smallest_two_rr])
    fifth_perc_rr = np.percentile(min_dist_rr,5)
    min_dist_ff = np.array([i[0] for i in smallest_two_ff])
    fifth_perc_ff = np.percentile(min_dist_ff,5)
    
    return np.array([fifth_perc_rf,fifth_perc_rr,fifth_perc_ff,nn_fifth_perc_rf,nn_fifth_perc_rr,nn_fifth_perc_ff]).reshape(1,6)    